var CustomPropsHandler = /** @class */ (function () {
    function CustomPropsHandler() {
    }
    /**
     * Register a new props handler
     *
     * @param handler to register for props (a handler can only be registered once per )
     * @returns a reference that can be used to unregister.
     */
    CustomPropsHandler.RegisterPropsHandler = function (propsHandler) {
        var propsChangeType = propsHandler.propChangeType;
        if (!Array.isArray(CustomPropsHandler._registeredPropsHandlers[propsChangeType])) {
            CustomPropsHandler._registeredPropsHandlers[propsChangeType] = [];
        }
        var registeredHandlers = CustomPropsHandler._registeredPropsHandlers[propsChangeType];
        var match = registeredHandlers.find(function (h) { return h === propsHandler; });
        if (match !== undefined) {
            console.warn("Handler can only be registered once per type [" + propsChangeType + "]");
            return match;
        }
        registeredHandlers.push(propsHandler);
        return propsHandler;
    };
    /**
     * Unregister a props handler that was previously registered.
     *
     * @param propsHandler
     *
     * @returns if the props handler was found and unregistered
     */
    CustomPropsHandler.UnregisterPropsHandler = function (propsHandlerToUnregister) {
        var propsChangeType = propsHandlerToUnregister.propChangeType;
        if (!Array.isArray(CustomPropsHandler._registeredPropsHandlers[propsChangeType])) {
            console.warn("cannot find " + propsHandlerToUnregister.name + " to unregister.");
            return false;
        }
        var registeredHandlers = CustomPropsHandler._registeredPropsHandlers[propsChangeType];
        var index = registeredHandlers.indexOf(propsHandlerToUnregister);
        if (index === -1) {
            console.warn("cannot find " + propsHandlerToUnregister.name + " to unregister.");
            return false;
        }
        CustomPropsHandler._registeredPropsHandlers[propsChangeType] = registeredHandlers.slice(index, 1);
        return true;
    };
    CustomPropsHandler.HandlePropsChange = function (propsChangeType, oldProp, newProp) {
        var registeredHandlers = CustomPropsHandler._registeredPropsHandlers[propsChangeType];
        if (registeredHandlers === undefined) {
            return CustomPropsHandler.NOT_ACCEPTED;
        }
        var accepted = false;
        for (var _i = 0, registeredHandlers_1 = registeredHandlers; _i < registeredHandlers_1.length; _i++) {
            var handler = registeredHandlers_1[_i];
            if (handler.accept(newProp)) {
                accepted = true;
                var propertyUpdatedProcessResult = handler.process(oldProp, newProp);
                // console.log(`handler '${handler.name}'custom prop processing result:`, oldProp, newProp, propertyUpdatedProcessResult);
                // give other custom handlers (if any) the opportunity to handle
                if (propertyUpdatedProcessResult.processed) {
                    return {
                        accepted: true,
                        processed: true,
                        value: propertyUpdatedProcessResult.value
                    };
                }
            }
        }
        return accepted
            ? CustomPropsHandler.ACCEPTED_NOT_PROCSSED
            : CustomPropsHandler.NOT_ACCEPTED;
    };
    CustomPropsHandler._registeredPropsHandlers = {};
    CustomPropsHandler.NOT_ACCEPTED = Object.freeze({
        accepted: false,
        processed: false,
        value: null
    });
    CustomPropsHandler.ACCEPTED_NOT_PROCSSED = Object.freeze({
        accepted: true,
        processed: false,
        value: null
    });
    return CustomPropsHandler;
}());
export { CustomPropsHandler };
export var PropChangeType;
(function (PropChangeType) {
    PropChangeType["Primitive"] = "Primitive";
    PropChangeType["Vector3"] = "Vector3";
    PropChangeType["Color3"] = "Color3";
    PropChangeType["Color4"] = "Color4";
    PropChangeType["Control"] = "Control";
    PropChangeType["NumericArray"] = "NumericArray";
    PropChangeType["Observable"] = "Observable";
    PropChangeType["Method"] = "Method";
    PropChangeType["LambdaExpression"] = "LambdaExpression";
    PropChangeType["FresnelParameters"] = "FresnelParameters";
    PropChangeType["Texture"] = "Texture";
})(PropChangeType || (PropChangeType = {}));
/**
 * Encapsulates common error handling and handling of registered custom prop handlers.
 */
function propertyCheck(oldProp, newProp, propertyName, propertyType, propChangeType, changedProps, templateMethod) {
    try {
        var processedResult = CustomPropsHandler.HandlePropsChange(propChangeType, oldProp, newProp);
        if (processedResult.accepted) {
            if (processedResult.processed) {
                // console.log(`handled ${propChangeType} on ${propertyName} by custom handler - new Value: ${JSON.stringify(processedResult.value ?? {})}`);
                changedProps.push({
                    propertyName: propertyName,
                    type: propertyType,
                    changeType: propChangeType,
                    value: processedResult.value
                });
            }
            // we assume any custom handler that 'accepts' a property does not want to pass
            // down to built-in handler.  ie: [] -> Vector3 would fail.
            return;
        }
        templateMethod(oldProp, newProp, changedProps);
    }
    catch (e) {
        console.error("Unable to update '" + propertyName + "' with " + propChangeType + ":", newProp);
        console.error(e);
    }
}
export var checkVector3Diff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    propertyCheck(oldProp, newProp, propertyName, propertyType, PropChangeType.Vector3, changedProps, function (oldProp, newProp, changedProps) {
        if (newProp && (!oldProp || !newProp.equals(oldProp))) {
            changedProps.push({
                propertyName: propertyName,
                type: propertyType,
                changeType: PropChangeType.Vector3,
                value: newProp
            });
        }
    });
};
export var checkColor3Diff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    propertyCheck(oldProp, newProp, propertyName, propertyType, PropChangeType.Color3, changedProps, function (oldProp, newProp, changedProps) {
        if (newProp && (!oldProp || !newProp.equals(oldProp))) {
            changedProps.push({
                propertyName: propertyName,
                type: propertyType,
                changeType: PropChangeType.Color3,
                value: newProp
            });
        }
    });
};
export var checkColor4Diff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    propertyCheck(oldProp, newProp, propertyName, propertyType, PropChangeType.Color4, changedProps, function (oldProp, newProp, changedProps) {
        // Color4.equals() not added until PR #5517
        if (newProp && (!oldProp || oldProp.r !== newProp.r || oldProp.g !== newProp.g || oldProp.b !== newProp.b || oldProp.a !== newProp.a)) {
            changedProps.push({
                propertyName: propertyName,
                type: propertyType,
                changeType: PropChangeType.Color4,
                value: newProp
            });
        }
    });
};
export var checkFresnelParametersDiff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    propertyCheck(oldProp, newProp, propertyName, propertyType, PropChangeType.FresnelParameters, changedProps, function (oldProp, newProp, changedProps) {
        // FresnelParameters.equals() not added until PR #7818 (https://github.com/BabylonJS/Babylon.js/pull/7818)
        if (newProp /* won't clear the property value */ && (!oldProp ||
            oldProp.bias !== newProp.bias ||
            oldProp.power !== newProp.power ||
            (!oldProp.leftColor || !oldProp.leftColor.equals(newProp.leftColor)) ||
            (!oldProp.rightColor || !oldProp.rightColor.equals(newProp.rightColor)) ||
            oldProp.isEnabled !== newProp.isEnabled)) {
            changedProps.push({
                propertyName: propertyName,
                type: propertyType,
                changeType: PropChangeType.FresnelParameters,
                value: newProp
            });
        }
    });
};
export var checkLambdaDiff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    propertyCheck(oldProp, newProp, propertyName, propertyType, PropChangeType.LambdaExpression, changedProps, function (oldProp, newProp, changedProps) {
        if (newProp !== oldProp) {
            changedProps.push({
                propertyName: propertyName,
                type: propertyType,
                changeType: PropChangeType.LambdaExpression,
                value: newProp
            });
        }
    });
};
export var checkControlDiff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    propertyCheck(oldProp, newProp, propertyName, propertyType, PropChangeType.Control, changedProps, function (oldProp, newProp, changedProps) {
        // only sets once
        if (newProp && !oldProp) {
            changedProps.push({
                propertyName: propertyName,
                type: propertyType,
                changeType: PropChangeType.Control,
                value: newProp
            });
        }
    });
};
export var checkPrimitiveDiff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    propertyCheck(oldProp, newProp, propertyName, propertyType, PropChangeType.Primitive, changedProps, function (oldProp, newProp, changedProps) {
        if (newProp !== oldProp) {
            changedProps.push({
                propertyName: propertyName,
                type: propertyType,
                changeType: PropChangeType.Primitive,
                value: newProp
            });
        }
    });
};
export var checkTextureDiff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    propertyCheck(oldProp, newProp, propertyName, propertyType, PropChangeType.Texture, changedProps, function (oldProp, newProp, changedProps) {
        if (newProp !== oldProp) {
            console.log('pushing texture:', propertyName, propertyType);
            changedProps.push({
                propertyName: propertyName,
                type: propertyType,
                changeType: PropChangeType.Texture,
                value: newProp
            });
        }
    });
};
export var checkNumericArrayDiff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    propertyCheck(oldProp, newProp, propertyName, propertyType, PropChangeType.NumericArray, changedProps, function (oldProp, newProp, changedProps) {
        // just length - missing loop + indexOf comparison (or deepEquals())
        if (newProp && (!oldProp || oldProp.length !== newProp.length)) {
            changedProps.push({
                propertyName: propertyName,
                type: propertyType,
                changeType: PropChangeType.NumericArray,
                value: newProp
            });
        }
    });
};
export var checkObservableDiff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    propertyCheck(oldProp, newProp, propertyName, propertyType, PropChangeType.Observable, changedProps, function (oldProp, newProp, changedProps) {
        // if it starts with 'on' then we have different handling.
        if (oldProp === undefined && oldProp !== newProp) {
            changedProps.push({
                propertyName: propertyName,
                type: propertyType,
                changeType: PropChangeType.Observable,
                value: newProp
            });
        }
    });
};
/**
 * This method differs from the rest in that we need to pass in a list of arguments.  Can be done by using:
 * 1. an array to signify each parameter (or Object.values(...)).
 *
 * @param oldProp value of method (array, object, primitive, etc.)
 * @param newProp value of method (array, object, primitive, etc.)
 * @param propertyName name of property for diff
 * @param propertyType signature of method (as string)
 * @param changedProps list to append to when a diff is found
 */
export var checkMethodDiff = function (oldProp, newProp, propertyName, propertyType, changedProps) {
    if (oldProp !== newProp) {
        changedProps.push({
            propertyName: propertyName,
            type: propertyType,
            changeType: PropChangeType.Method,
            value: newProp
        });
    }
};
//# sourceMappingURL=PropsHandler.js.map